(* Title:  BaseDefinitions/theoryC.sml
   Author: Walther Neuper
   (c) due to copyright terms

Probably this structure can be dropped due to improved reflection in Isabelle.
*)
signature THEORY_ISAC =
sig
  eqtype id
  val cut_longid: string -> id
  val id_of: theory -> id
  val get_theory: Proof.context -> id -> theory

  val id_empty: id
  val curried_equal: theory -> theory -> bool
  val parent_of: theory -> theory -> theory

  type metadata
  val the_metadata: theory -> metadata
  val set_metadata: metadata -> theory -> theory

\<^isac_test>\<open>
(**)
\<close>
end

(**)
structure ThyC(**): THEORY_ISAC(**) =
struct
(**)

type id = string;
fun cut_longid dn = hd (space_explode "." dn);
fun id_of thy = thy |> Context.theory_id |> Context.theory_id_name

fun get_theory ctxt thy_id =
  let
    val known_thys = Theory.nodes_of (Proof_Context.theory_of ctxt)
  in 
    (find_first (fn thy => Context.theory_name thy = thy_id) known_thys
      |> the)
    handle Option.Option => raise ERROR ("get_theory fails with " ^ quote thy_id)
  end

val id_empty = "empty_thy_id";
fun curried_equal thy1 thy2 = id_of thy1 = id_of thy2
fun parent_of thy1 thy2 = if Context.subthy (thy1, thy2) then thy2 else thy1;


(** meta data **)

type authors = string list; (* no more for thm, rls, rew_ord, cal *)
type metadata = {coursedesign: authors, mathauthors: authors}
(*WAS IN Thy_Write:  datatype thydata
    = ...
    | Html of {coursedesign: authors, guh: Check_Unique.id, html: string, mathauthors: authors}*)
val default_metadata = {coursedesign = ["(c) isac team 2006"], mathauthors = []}

structure Data = Theory_Data
(
  type T = metadata option;
  val empty = SOME default_metadata;
  fun merge _ = NONE;
);
val the_metadata = the o Data.get;
val set_metadata = Data.put o SOME;


(**)end(**)
