(* Title:  BaseDefinitions/unparseC.sml
   Author: Walther Neuper 2018
   (c) copyright due to lincense terms
*)

signature UNPARSE_ISAC =
sig
  eqtype term_as_string

  val term_empty: term
  val typ_empty: typ

  val term: Proof.context -> term -> term_as_string
(*val term_ERROR \<longrightarrow> UnparseC.term (ContextC.for_ERROR ()): term -> term_as_string, due to file dependencies*)
  val term_in_thy: theory -> term -> term_as_string
  val term_opt: Proof.context -> term option -> term_as_string

  val terms: Proof.context -> term list -> term_as_string
  val terms_in_thy: theory -> term list -> term_as_string
  val typ: Proof.context -> typ -> term_as_string

\<^isac_test>\<open>
  val asms_test: Proof.context -> term list -> term_as_string list
\<close>
end

(**)
structure UnparseC(**): UNPARSE_ISAC(**) =
struct
(**)

type term_as_string = string;

fun term ctxt t =
  let
    val ctxt' = Config.put show_markup false ctxt;
  in
    Print_Mode.setmp [] (Syntax.string_of_term ctxt') t
  end;
fun terms ctxt ts = ts |> map (term ctxt) |> strs2str';

fun term_in_thy thy t = term (Proof_Context.init_global thy) t;
fun terms_in_thy thy ts = ts |> map (term_in_thy thy) |> strs2str';

fun typ ctxt typ = Print_Mode.setmp [] (Syntax.string_of_typ ctxt) typ

fun term_opt ctxt (SOME t) = "(SOME " ^ term ctxt t ^ ")"
  | term_opt _ NONE = "NONE";

val typ_empty = Type ("empty", []);
val term_empty = Const ("empty", Type("'a", []));

\<^isac_test>\<open>
(*for eq_set, in particular for (re-ordered) assumptions*)
fun asms_test ctxt ts = map (term ctxt) ts;
\<close>

(**)end(**)
