(*  Title:  $ISABELLE_ISAC_TEST/Pure/General/Basics.thy
    Author: Walther Neuper, TU Graz, 101027
*)

theory Basics
imports Complex_Main
begin

subsection \<open>Pure/General/basics.ML line by line\<close>
text \<open>
  Combinators are best understood by looking at their fun definition
  in relation to their signature.
\<close>
subsubsection \<open>application and structured results\<close>
(*fun       x |>     f       = f x;
  val  |> : 'a * ('a -> 'b) -> 'b*)
ML \<open>op |> ;
  1 |> (fn x => x + 10)
    |> (fn x => 2 * x);
(*val it = 22 : int*)
\<close>
(*fun        (x  , y) |->    f             = f x y;
  val  |-> : ('c * 'a) * ('c -> 'a -> 'b) -> 'b*)
ML \<open>op |-> ;
  fun ff c a = a + c;
  (10, 1) |-> ff
(*val it = 11 : int*)
\<close>
(*fun        (x  , y) |>>    f       = (f x, y);
  val  |>> : ('a * 'c) * ('a -> 'b) -> 'b * 'c*)
ML \<open>op |>> ;
  (1, "a") |>> (fn x => x + 10)
(*val it = (11, "a") : int * string*)
\<close>
(*fun        (x  , y) ||>    f       = (x, f y);
  val  ||> : ('c * 'a) * ('a -> 'b) -> 'c * 'b*)
ML \<open>op ||> ;
  ("a", 1) ||> (fn x => x + 10)
(*val it = ("a", 11) : string * int*)
\<close>
(*fun           (x , y) ||>>    f = let val (z, y') = f y in ((x, z), y') end;
  val   ||>> : ('c * 'a) * ('a -> 'd * 'b)        ->     ('c * 'd) * 'b
          applied to ^^                  and re-arranged ^^^    ^^^^^^^*)
ML \<open> op ||>>;
  (444, 2) ||>> (fn e => (e * 3, e + 5));
(*val it = ((444, 6), 7): (int * int) * int*)
\<close>

subsubsection \<open>composition and structured results\<close>
text \<open>
  Note the analogies between |* above and #* below.
\<close>
ML \<open>
  op #>;
  ((fn a => a + 10) #> (fn a => a + 5)) 3;
  (*val it = 18: int*)
\<close>

ML \<open>
  op #->;
  fun ff c a = a + c;
  ((fn e => (e * 2, e + 99))  #-> ff) 50;
  (*val it = 249: int*)
\<close>

ML \<open>
  op #>>;
  fun fff a = (a + 10, string_of_int a);
  fun ggg b = b * 2;
  fff #>> ggg;
  (fff #>> ggg) 1;
  (*val it = (22, "1"): int * string*)
\<close>

ML \<open>
  op ##>;
  ((fn e => (e * 3, e + 4)) ##> (fn a => a + 5)) 6;
  (*val it = (18, 15): int * int*)
\<close>

ML \<open>
  op ##>>;
  ((fn e => (e * 2, e + 3)) ##>> (fn u => (u * 4, u + 5))) 6;
  (*val it = ((12, 36), 14): (int * int) * int*)
\<close>

subsection \<open>Decompose chains of combinators\<close>
text \<open>
  This is an example taken from fun parse_vcs
  in HOL/SPARK/Tools/fdl_parser.ML
\<close>
ML \<open>
fun decompose_combinators str last_fun =
(*1*)str |>
(*2*)pair ("snd-" ^ str) ||>
(*3*)(fn str => (str ^ ".1")) ||>
(*4*)pair (fn str => (str ^ ".2")) ||>
(*5*)last_fun
\<close> ML \<open>
case decompose_combinators "abc" fst of
  ("snd-abc", _): string * (string -> string) => ()
  | _ => error "decompose_combinators fst CHANGED";
if decompose_combinators "abc" snd = ("snd-abc", "abc.1") then ()
  else error "decompose_combinators snd CHANGED";
\<close> ML \<open>
fun decomposed_combinators str last_fun =
  let
    val line_2 = pair ("snd-" ^ str) str
    val line_3 = apsnd (fn str => (str ^ ".1")) line_2
    val line_4 = apsnd (pair (fn str => (str ^ ".1"))) line_3
    val line_5 = apsnd last_fun line_4
  in line_5 end
\<close> ML \<open>
case decomposed_combinators "abc" fst of
  ("snd-abc", _): string * (string -> string) => ()
  | _ => error "decomposed_combinators fst CHANGED";
if decomposed_combinators "abc" snd = ("snd-abc", "abc.1") then ()
  else error "decomposed_combinators snd CHANGED";
\<close> ML \<open>
\<close>


end

